#include <string>
#include <vector>
#include "stoch.hh"
//#include <climits>
#include <float.h>

prioriQ::prioriQ (reactionDataTable& rReactionDataTable, int index)
{
  totalTime = 0;
  sizeTimeArray = rReactionDataTable.reactionDataVector.size();
  timeArray = new reactionData*[sizeTimeArray];
  for (int i = 0; i < sizeTimeArray; i++)
    {
      timeArray[i] = rReactionDataTable.reactionDataVector[i];
      rReactionDataTable.reactionDataVector[i]->times[index]->timeArrayPosition = i;
      if (i != 0)
	{
	  bubbleUp(i, index);
	  
	}
    }
}

void 
prioriQ::switchTimes (const int p, const int q, int index)
{
  timeArray[p]->times[index]->timeArrayPosition = q;
  timeArray[q]->times[index]->timeArrayPosition = p;
  reactionData* tmp = timeArray[p];
  timeArray[p] = timeArray[q];
  timeArray[q] = tmp;
}

int
prioriQ::findParent (int node)
{
  if ((node % 2) == 0)
    return (node-2)/2;
  else
    return (node-1)/2;
}
  
void
prioriQ::bubbleUp (int node, int index)
{
  if (node != 0)
    {
      int parent = findParent(node);
      if ((timeArray[node]->times[index]->time) < (timeArray [parent]->times[index]->time))
	{
	  switchTimes (node, parent, index);
	  node = parent;
	  bubbleUp (node, index);
	}
    }
}

int
prioriQ::childCheck (int node)
{
  if ((2*node+2) <= (sizeTimeArray-1))
    return 2;
  else if ((2*node+1) == (sizeTimeArray-1))
    return 1;
  else
    return 0;
}

void
prioriQ::bubbleDown (int node, int index)
{
  int children = childCheck (node);
  switch (children)
    {
    case 2:
      // check if lt. child is < rt.
      if (timeArray[(2*node+1)]->times[index]->time < timeArray[(2*node+2)]->times[index]->time) 
	{
	  if (timeArray[(2*node+1)]->times[index]->time < timeArray[node]->times[index]->time)
	    {
	      switchTimes ((2*node+1), node, index);
	      node = (2*node+1);
	      bubbleDown (node, index);  // do it again
	    }
	}
      else // rc is < lc
	{
	  if (timeArray[(2*node+2)]->times[index]->time < timeArray[node]->times[index]->time)
	    {
	      switchTimes ((2*node +2), node, index );
	      node  = (2*node +2);
	      bubbleDown (node,index);
	    }
	}
      break;
    case 1:
      if (timeArray[(2*node +1)]->times[index]->time < timeArray[node]->times[index]->time) 
	{
	  switchTimes ((2*node+1), node, index);
	  node = (2*node+1);
	}
      break;
    case 0:
      break;
    }
}

void 
prioriQ::dumpPrioriQ (int index, ostream& outFile)
{
  
  outFile << "Time\t\t\thc\t\t\tName" << endl;
  for (int i = 0; i < sizeTimeArray; i++)
    {
      outFile.width(10);
      outFile << timeArray[i]->times[index]->time; 
      outFile.width(10);
      outFile << "\t" << timeArray[i]->times[index]->hc;
      outFile.width(10);
      outFile << "\t" << timeArray[i]->reactionName << endl;
      timeArray[i]->dumpType(outFile, index);
      timeArray[i]->times[index]->dumpTimeStuff(outFile);
      outFile << endl;
      
    }
  outFile << "Total time:  " << totalTime;
  outFile << endl << endl;
}

void 
prioriQ::dumpPrioriQ0 (int index, ostream& outFile)
{
  
  outFile << "Name:  " << timeArray[0]->reactionName << endl;
  timeArray[0]->dumpType(outFile, index);
  timeArray[0]->times[index]->dumpTimeStuff(outFile); 
  outFile << "Total time:  " << totalTime;
  outFile << endl << endl;
}

void
prioriQ::bubble (int node, int index)
{
  if (node != 0) // not root
    {
      if (timeArray[node]->times[index]->timeInfinity)
      	{
      	  bubbleDown(node, index);
      	  return;
      	}
      int parent = findParent(node);
      
      if (timeArray[node]->times[index]->time < timeArray[parent]->times[index]->time) 
	bubbleUp (node, index);
      else bubbleDown (node, index);
    }
  else
    bubbleDown (node, index);
}
    


void
prioriQ::executeReaction (int node, int index, ostream& outFile)
{
  totalTime = timeArray[node]->times[index]->time;
  //updatePQ (timeArray[node]->times[index]->time, index);
  timeArray[node]->decreaseSubstrates(index, totalTime, outFile);
  timeArray[node]->increaseProducts(index, totalTime, outFile);
  timeArray[node]->times[index]->calculateNewTime(timeArray[node]->hcEqu(index),totalTime); // calculate a new time for root
  reactionData* tmp =  timeArray[node];//->pReactionGraph[0];
  bubble (node, index); // sort.  Note:  root of pQ doesn't point to executed rxn any more.
  // Starts at the second element of the reaction graph 'cause the first 
  // element is self.  May think about removing from reaction graph if not 
  // ever used.
  for (int q = 1; q < tmp->pReactionGraph.size(); q++)
    {
      if (tmp->pReactionGraph[q]->times[index]->timeInfinity)
	{
	  tmp->pReactionGraph[q]->times[index]->calculateNewTime(tmp->pReactionGraph[q]->hcEqu(index),totalTime);
	}
      else
	{
	  tmp->pReactionGraph[q]->times[index]->updateTime(tmp->pReactionGraph[q]->hcEqu(index),totalTime); 
	}
      bubble(tmp->pReactionGraph[q]->times[index]->timeArrayPosition, index);
    }
}   

void
prioriQ::updatePQ (double time, int index)
{

  for (int i = 0; i < sizeTimeArray; i++)
    {
      if (!timeArray[i]->times[index]->timeInfinity)
	timeArray[i]->times[index]->time = timeArray[i]->times[index]->time - time;
    }
}


  


prioriQtable::prioriQtable (int inReplicate, reactionDataTable& rReactionDataTable)
{
  cout << "Initializing priority queue table . . .";
  replicate = inReplicate;
  for (int i = 0; i < replicate; i++)
    {
      pqVector.push_back( new prioriQ(rReactionDataTable, i));
    }
  cout << "Finished!" << endl << endl;
}

 prioriQtable::~prioriQtable(void)
{
  for (int i = 0; i < pqVector.size(); i++)
    {
      
      delete pqVector[i];
      
    }
}


void 
prioriQtable::dumpPrioriQtable (ostream& outFile)
{
  outFile << "Priori Q table:" << endl;
  for (int p = 0; p < replicate; p++)
    {
      outFile << "\tReplicate " << p << endl;
      pqVector[p]->dumpPrioriQ(p, outFile);
    }
}










