#include <fstream>
#include <string>
#include <vector>
#include "stoch.hh"
#include <cmath>
#include <float.h>

using namespace std;

reactionData::reactionData(const string inReactionName, 
			   vector <string> inSubstrates, 
	      vector <string> inProducts, 
	      const float inReactionRate, 
	      const speciesTable& specTab) :
  reactionName(inReactionName),
  reactionRate(inReactionRate)

{
  for (int n = 0; n < inSubstrates.size(); n++) 
    pSubstrateSpecies.push_back
      (specTab.speciesLookup(inSubstrates[n]));
  for (int n = 0; n < inProducts.size(); n++)
    pProductSpecies.push_back
      (specTab.speciesLookup(inProducts[n]));
}

reactionData::~reactionData(void)
{
  for (int i =0; i < times.size(); i++)
    {
      delete times[i];
    }
}

// equal number method

//  void
//  reactionData::decreaseSubstrates (const int rep, const double totalTime, ostream& outFile)
//  {
//    for (int i = 0; i < pSubstrateSpecies.size(); i++)
//      {
//        if (pSubstrateSpecies[i]->aAmountSpecies[rep] != 0)
//  	{
//  	  pSubstrateSpecies[i]->aAmountSpecies[rep]--;
//  	  if (pSubstrateSpecies[i]->comp)
//  	    {
//  	      if (pSubstrateSpecies[i]->aAmountSpecies[rep] == pSubstrateSpecies[i]->compSpecies->
//  		  aAmountSpecies[rep])
//  		{
//  		  outFile << pSubstrateSpecies[i]->speciesName << "-" 
//  			  << pSubstrateSpecies[i]->compSpecies->speciesName
//  			  << "-equi"
//  			  << "\t" << totalTime 
//  			  << "\t" << (totalTime-pSubstrateSpecies[i]->lastTime) 
//  			  << "\t" << pSubstrateSpecies[i]->aAmountSpecies[rep]
//  			  << endl; 
//  		  pSubstrateSpecies[i]->lastTime = totalTime;
//  		}
//  	    }
//  	  if (pSubstrateSpecies[i]->comped)
//  	    {
//  	      if (pSubstrateSpecies[i]->aAmountSpecies[rep] == pSubstrateSpecies[i]->compedSpecies->
//  		  aAmountSpecies[rep])
//  		{
//  		  outFile << pSubstrateSpecies[i]->compedSpecies->speciesName << "-" 
//  			  << pSubstrateSpecies[i]->speciesName
//  			  << "-equi"
//  			  << "\t" << totalTime 
//  			  << "\t" << (totalTime-pSubstrateSpecies[i]->compedSpecies->lastTime) 
//  			  << "\t" << pSubstrateSpecies[i]->aAmountSpecies[rep]
//  			  << endl; 
//  		  pSubstrateSpecies[i]->compedSpecies->lastTime = totalTime;
//  		}
//  	    }
//  	}
//        else 
//  	{
//  	  cout << "There is no more of species " 
//  	       << pSubstrateSpecies[i]->speciesName
//  	       << " left!" << endl;
//  	  cout << reactionName << " hc:  " << times[rep]->hc;
//  	  cout << " time " << times[rep]->time;
//  	  cout << "inf flag = " << times[rep]->timeInfinity << endl;
//  	  string tmp;
//  	  cin >> tmp;
//  	}
//      }
//  }

//  void
//  reactionData::increaseProducts (const int rep, const double totalTime, ostream& outFile)
//  {
//    for (int i = 0; i < pProductSpecies.size(); i++)
//      {
//        pProductSpecies[i]->aAmountSpecies[rep]++;   
//        if (pProductSpecies[i]->comp)
//  	{
//  	  if (pProductSpecies[i]->aAmountSpecies[rep] == pProductSpecies[i]->compSpecies->
//  	      aAmountSpecies[rep])
//  	    {
//  	      outFile << pProductSpecies[i]->speciesName << "-" 
//  		      << pProductSpecies[i]->compSpecies->speciesName 
//  		      << "-equi"
//  		      << "\t" << totalTime 
//  		      << "\t" << (totalTime-pProductSpecies[i]->lastTime) 
//  		      << "\t" << pProductSpecies[i]->aAmountSpecies[rep]
//  		      << endl; 
//  	      pProductSpecies[i]->lastTime = totalTime;
//  	    }
//  	}
      
//        if (pProductSpecies[i]->comped)
//  	{
//  	  if (pProductSpecies[i]->aAmountSpecies[rep] == pProductSpecies[i]->compedSpecies->
//  	      aAmountSpecies[rep])
//  	    {
//  	      outFile << pProductSpecies[i]->compedSpecies->speciesName << "-" 
//  		      << pProductSpecies[i]->speciesName 
//  		      << "-equi"
//  		      << "\t" << totalTime 
//  		      << "\t" << (totalTime-pProductSpecies[i]->compedSpecies->lastTime) 
//  		      << "\t" << pProductSpecies[i]->aAmountSpecies[rep]
//  		      << endl; 
//  	      pProductSpecies[i]->compedSpecies->lastTime = totalTime;
//  	    }
//  	}
//      }
//  }


// greater than less than method


void
reactionData::decreaseSubstrates (const int rep, const double totalTime, ostream& outFile)
{
  for (int i = 0; i < pSubstrateSpecies.size(); i++)
    {
      if (pSubstrateSpecies[i]->aAmountSpecies[rep] != 0)
	{
	  pSubstrateSpecies[i]->aAmountSpecies[rep]--;
	  if (pSubstrateSpecies[i]->comp)
	    {
	      if (pSubstrateSpecies[i]->aAmountSpecies[rep] < pSubstrateSpecies[i]->compSpecies->
		  aAmountSpecies[rep] && pSubstrateSpecies[i]->vGreater[rep])
		{
		  pSubstrateSpecies[i]->vGreater[rep] = 0;
		  outFile << pSubstrateSpecies[i]->speciesName << "-" 
			  << pSubstrateSpecies[i]->compSpecies->speciesName 
			  << "-gtlt"
			  << "\t" << totalTime 
			  << "\t" << (totalTime-pSubstrateSpecies[i]->lastTimeGtlt) 
			  << "\t" << pSubstrateSpecies[i]->aAmountSpecies[rep]
		          << "\t" << pSubstrateSpecies[i]->compSpecies->aAmountSpecies[rep]
                          << endl; 
		  pSubstrateSpecies[i]->lastTimeGtlt = totalTime;
		  
		}
	      else if (pSubstrateSpecies[i]->aAmountSpecies[rep] > pSubstrateSpecies[i]->
		       compSpecies->aAmountSpecies[rep] && !pSubstrateSpecies[i]->vGreater[rep])
		{
		  pSubstrateSpecies[i]->vGreater[rep] = 1;
		  outFile << pSubstrateSpecies[i]->speciesName << "-" 
			  << pSubstrateSpecies[i]->compSpecies->speciesName 
			  << "-gtlt"
			  << "\t" << totalTime 
			  << "\t" << (totalTime-pSubstrateSpecies[i]->lastTimeGtlt) 
			  << "\t" << pSubstrateSpecies[i]->aAmountSpecies[rep]
		          << "\t" << pSubstrateSpecies[i]->compSpecies->aAmountSpecies[rep]
                          << endl; 
		  pSubstrateSpecies[i]->lastTimeGtlt = totalTime;
		}
	      //equal number method
	      else if (pSubstrateSpecies[i]->aAmountSpecies[rep] == pSubstrateSpecies[i]->compSpecies->
		  aAmountSpecies[rep])
		{
		  outFile << pSubstrateSpecies[i]->speciesName << "-" 
			  << pSubstrateSpecies[i]->compSpecies->speciesName
			  << "-equi"
			  << "\t" << totalTime 
			  << "\t" << (totalTime-pSubstrateSpecies[i]->lastTimeEqui) 
			  << "\t" << pSubstrateSpecies[i]->aAmountSpecies[rep]
			  << endl; 
		  pSubstrateSpecies[i]->lastTimeEqui = totalTime;
		}
	    }
	  if (pSubstrateSpecies[i]->comped)
	    {
	      if (pSubstrateSpecies[i]->compedSpecies->aAmountSpecies[rep] < 
		  pSubstrateSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep] && 
		  pSubstrateSpecies[i]->compedSpecies->vGreater[rep])
		{
		  pSubstrateSpecies[i]->compedSpecies->vGreater[rep] = 0;
		  outFile << pSubstrateSpecies[i]->compedSpecies->speciesName << "-" 
			  << pSubstrateSpecies[i]->compedSpecies->compSpecies->speciesName 
			  << "-gtlt"
			  << "\t" << totalTime 
			  << "\t" << (totalTime-pSubstrateSpecies[i]->compedSpecies->lastTimeGtlt) 
			  << "\t" << pSubstrateSpecies[i]->compedSpecies->aAmountSpecies[rep]
		          << "\t" << pSubstrateSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep]
                          << endl; 
		  pSubstrateSpecies[i]->compedSpecies->lastTimeGtlt = totalTime;
		  
		}
	      else if (pSubstrateSpecies[i]->compedSpecies->aAmountSpecies[rep] > 
		       pSubstrateSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep] && 
		       !pSubstrateSpecies[i]->compedSpecies->vGreater[rep])
		{
		  pSubstrateSpecies[i]->compedSpecies->vGreater[rep] = 1;
		  outFile << pSubstrateSpecies[i]->compedSpecies->speciesName << "-" 
			  << pSubstrateSpecies[i]->compedSpecies->compSpecies->speciesName 
			  << "-gtlt"
			  << "\t" << totalTime 
			  << "\t" << (totalTime-pSubstrateSpecies[i]->compedSpecies->lastTimeGtlt) 
			  << "\t" << pSubstrateSpecies[i]->compedSpecies->aAmountSpecies[rep]
		          << "\t" << pSubstrateSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep]
                          << endl; 
		  pSubstrateSpecies[i]->compedSpecies->lastTimeGtlt = totalTime;
		}
	      // equal number method
	      else if (pSubstrateSpecies[i]->aAmountSpecies[rep] == pSubstrateSpecies[i]->compedSpecies->
		  aAmountSpecies[rep])
		{
		  outFile << pSubstrateSpecies[i]->compedSpecies->speciesName << "-" 
			  << pSubstrateSpecies[i]->speciesName
			  << "-equi"
			  << "\t" << totalTime 
			  << "\t" << (totalTime-pSubstrateSpecies[i]->compedSpecies->lastTimeEqui) 
			  << "\t" << pSubstrateSpecies[i]->aAmountSpecies[rep]
			  << endl; 
		  pSubstrateSpecies[i]->compedSpecies->lastTimeEqui = totalTime;
		}
	    }
	}
      else 
	{
	  cout << "There is no more of species " 
	       << pSubstrateSpecies[i]->speciesName
	       << " left!" << endl;
	  cout << reactionName << " hc:  " << times[rep]->hc;
	  cout << " time " << times[rep]->time;
	  cout << "inf flag = " << times[rep]->timeInfinity << endl;
	  string tmp;
	  cin >> tmp;
	}
    }
}

void
reactionData::increaseProducts (const int rep, const double totalTime, ostream& outFile)
{
  for (int i = 0; i < pProductSpecies.size(); i++)
    {
      pProductSpecies[i]->aAmountSpecies[rep]++;   
      if (pProductSpecies[i]->comp)
	{
	  if (pProductSpecies[i]->aAmountSpecies[rep] < pProductSpecies[i]->compSpecies->
	      aAmountSpecies[rep] && pProductSpecies[i]->vGreater[rep])
	    {
	      pProductSpecies[i]->vGreater[rep] = 0;
	      outFile << pProductSpecies[i]->speciesName << "-" 
		      << pProductSpecies[i]->compSpecies->speciesName 
		      << "-gtlt"
		      << "\t" << totalTime 
		      << "\t" << (totalTime-pProductSpecies[i]->lastTimeGtlt) 
		      << "\t" << pProductSpecies[i]->aAmountSpecies[rep]
		      << "\t" << pProductSpecies[i]->compSpecies->aAmountSpecies[rep]
		      << endl; 
	      pProductSpecies[i]->lastTimeGtlt = totalTime;
	    }
	  else if (pProductSpecies[i]->aAmountSpecies[rep] > pProductSpecies[i]->
		   compSpecies->aAmountSpecies[rep] && !pProductSpecies[i]->vGreater[rep])
	    {
	      pProductSpecies[i]->vGreater[rep] = 1;
	      outFile << pProductSpecies[i]->speciesName << "-" 
		      << pProductSpecies[i]->compSpecies->speciesName 
		      << "-gtlt"
		      << "\t" << totalTime 
		      << "\t" << (totalTime-pProductSpecies[i]->lastTimeGtlt) 
		      << "\t" << pProductSpecies[i]->aAmountSpecies[rep]
		      << "\t" << pProductSpecies[i]->compSpecies->aAmountSpecies[rep]
		      << endl;
	      pProductSpecies[i]->lastTimeGtlt = totalTime;
	    }
	  //equal number method
	  else if (pProductSpecies[i]->aAmountSpecies[rep] == pProductSpecies[i]->compSpecies->
	      aAmountSpecies[rep])
	    {
	      outFile << pProductSpecies[i]->speciesName << "-" 
		      << pProductSpecies[i]->compSpecies->speciesName 
		      << "-equi"
		      << "\t" << totalTime 
		      << "\t" << (totalTime-pProductSpecies[i]->lastTimeEqui) 
		      << "\t" << pProductSpecies[i]->aAmountSpecies[rep]
		      << endl; 
	      pProductSpecies[i]->lastTimeEqui = totalTime;
	    }
	}
      if (pProductSpecies[i]->comped)
	{
	  if (pProductSpecies[i]->compedSpecies->aAmountSpecies[rep] < 
	      pProductSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep] && 
	      pProductSpecies[i]->compedSpecies->vGreater[rep])
	    {
	      pProductSpecies[i]->compedSpecies->vGreater[rep] = 0;
	      outFile << pProductSpecies[i]->compedSpecies->speciesName << "-" 
		      << pProductSpecies[i]->compedSpecies->compSpecies->speciesName 
		      << "-gtlt"
		      << "\t" << totalTime 
		      << "\t" << (totalTime-pProductSpecies[i]->compedSpecies->lastTimeGtlt) 
		      << "\t" << pProductSpecies[i]->compedSpecies->aAmountSpecies[rep]
		      << "\t" << pProductSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep]
		      << endl; 
	      pProductSpecies[i]->compedSpecies->lastTimeGtlt = totalTime;
	      
	    }
	  else if (pProductSpecies[i]->compedSpecies->aAmountSpecies[rep] > 
		   pProductSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep] && 
		   !pProductSpecies[i]->compedSpecies->vGreater[rep])
	    {
	      pProductSpecies[i]->compedSpecies->vGreater[rep] = 1;
	      outFile << pProductSpecies[i]->compedSpecies->speciesName << "-" 
		      << pProductSpecies[i]->compedSpecies->compSpecies->speciesName 
		      << "-gtlt"
		      << "\t" << totalTime 
		      << "\t" << (totalTime-pProductSpecies[i]->compedSpecies->lastTimeGtlt) 
		      << "\t" << pProductSpecies[i]->compedSpecies->aAmountSpecies[rep]
		      << "\t" << pProductSpecies[i]->compedSpecies->compSpecies->aAmountSpecies[rep]
		      << endl; 
	      pProductSpecies[i]->compedSpecies->lastTimeGtlt = totalTime;
	    }
	  // equal number method
	  else if (pProductSpecies[i]->aAmountSpecies[rep] == pProductSpecies[i]->compedSpecies->
	      aAmountSpecies[rep])
	    {
	      outFile << pProductSpecies[i]->compedSpecies->speciesName << "-" 
		      << pProductSpecies[i]->speciesName 
		      << "-equi"
		      << "\t" << totalTime 
		      << "\t" << (totalTime-pProductSpecies[i]->compedSpecies->lastTimeEqui) 
		      << "\t" << pProductSpecies[i]->aAmountSpecies[rep]
		      << endl; 
	      pProductSpecies[i]->compedSpecies->lastTimeEqui = totalTime;
	    }
	}
    }
}

reactionDataTable::~reactionDataTable(void)
{
  for(int i = 0; i < reactionDataVector.size(); i++)
    {
      delete reactionDataVector[i];
    }
}

void 
reactionDataTable::addNewReaction(reactionData* pReactionData)
{
  reactionDataVector.push_back(pReactionData);
}

void
reactionDataTable::readReactionDataFile(istream& rxnIn, 
					const speciesTable& rSpecTab)
{
  cout << "Beginning to read the reaction data file . . ." << endl;
  string evalLine;
  string rxnName;
  vector <string> subs;
  vector <string> prods;
  float rate;
  bool Subs = 0;
  bool Prods = 0;
  
  int type;
  while (!rxnIn.eof())
    {	
      
      rxnIn >> evalLine;
      if (evalLine == "Rxn") 
	{
	  rxnIn >> rxnName;
	  continue;
	}
      if (evalLine == "Subs") 
	{
	  Subs = 1;
	  continue;
	}
      if (evalLine == "Prods") 
	{
	  Prods = 1;
	  Subs = 0;
	  continue;
	}
      if (evalLine == "Rate") 
	{
	  Prods = 0;
	  rxnIn >> rate;
	  type = getType(subs);
	  evalLine = "";
	  switch (type)
	    {
	    case 0:
	      addNewReaction (new reactionTypeA (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;

	    case 1:
	      addNewReaction (new reactionTypeB (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case 2:
	      addNewReaction (new reactionTypeC (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case 3:
	      addNewReaction (new reactionTypeD (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case 4:
	      addNewReaction (new reactionTypeE (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case 5:
	      addNewReaction (new reactionTypeFa (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case 6:
	      addNewReaction (new reactionTypeFb (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case 7:
	      addNewReaction (new reactionTypeG (rxnName, subs, prods, 
						 rate, rSpecTab));
	      subs.clear();
	      prods.clear();
	      break;
	    case -1:
	      cout << endl << "Error parsing reaction \"" << rxnName;
	      cout << "\".  Unknown reaction type!" << endl;
	      cout << "Substrates:  ";
	      for (int i=0; i< subs.size(); i++)
		{
		  cout << subs[i] << " ";
		}
	      cout << endl;
	      exit (0);
	      break;
	    }
	    
	  continue;
	}
      // Prods evaluation must come after Rate evaluation otherwise prods 
      // stays true and gets "Rate" assigned to it as a final product
      if (Subs) subs.push_back(evalLine);
      if (Prods) prods.push_back(evalLine);
    }
  cout << "Finished!" << endl << endl;
}

void
reactionDataTable::dumpReactionDataTable (ostream& outFile)
{
  for (int q = 0; q < reactionDataVector.size(); q++)
    {
      outFile << "Name: " <<  reactionDataVector[q]->reactionName
	   << "\n\tsubstrates: ";
      for (int w =0; w < (reactionDataVector[q]->pSubstrateSpecies.size()); w++) 
	{
	  outFile << reactionDataVector[q]->pSubstrateSpecies[w]->speciesName 
	       << " ";
	}
      outFile << "\n\tproducts: ";
      for (int w= 0 ; w < (reactionDataVector[q]->pProductSpecies.size()); w++)
	outFile << reactionDataVector[q]->pProductSpecies[w]->speciesName << " ";
      outFile << "\n\trate: " << reactionDataVector[q]->reactionRate 
	      << "\n\ttype: ";
      reactionDataVector[q]->dumpType(outFile,0);
      
      outFile << endl;
      outFile << "\treaction graph: ";
      for (int w= 0 ; w < (reactionDataVector[q]->pReactionGraph.size()); w++)
	outFile << reactionDataVector[q]->pReactionGraph[w]->reactionName << " ";
      outFile << endl;
      for (int w = 0; w < reactionDataVector[q]->times.size(); w++)
	{
	  outFile << "\treplicate:  " << w << endl;
	  outFile << "\t\t";
	  reactionDataVector[q]->times[w]->dumpTimeStuff(outFile); 
	  outFile<< endl;
	}

      outFile << endl << endl;
    }
}

int
reactionDataTable::getType (vector <string> subs)
{
  int subsIndex = subs.size();
  if (subsIndex == 0) return 0;
  if (subsIndex == 1) return 1;
  if (subsIndex == 2) 
    { 
      if (subs[0] == subs [1]) return 3;
      else return 2; 
    }
  if (subsIndex == 3) 
    {
      if ((subs[0] != subs[1])  && (subs[0] != subs[2]) && (subs[1] != subs[2]))
	return 4; // a,b,c
      else if ((subs[0] == subs[1])  && (subs[0] == subs[2])) return 7; // a,a,a
      else if (subs[0] == subs[1]) 
	return 5; // a,a,b
      else
	return 6; // a,b,a or b,a,a
    }
  else 
    {
      return -1;
    }
}

void 
reactionDataTable::makeReactionGraph ()
{
  cout << "Starting to make the Reaction Graph. . .";
  
  bool exists;
  for (int i = 0; i < (reactionDataVector.size()); i++)
    {
      vector <species*> pSpeciesList;
      // The first thing to do is build a list of species that will be checked 
      // for use as substrates in other reactions.
      // Check for substrate in the products list, if found, don't include.
      // For the reaction, A+E->B+E, I don't want to include reactions containing 
      // E in the reaction graph.
     
      for (int q = 0; q < (reactionDataVector[i]->pSubstrateSpecies.size()); q++)
	{
	  bool ignoreSpecies = 0;
	  for (int e = 0; e < (reactionDataVector[i]->pProductSpecies.size()); e++)
	    {
	      if (reactionDataVector[i]->pSubstrateSpecies[q] == reactionDataVector[i]->pProductSpecies[e])
		{
		  ignoreSpecies = 1;
		  break;
		}
	    }
	  if (!ignoreSpecies)
	    {
	      pSpeciesList.push_back (reactionDataVector[i]->pSubstrateSpecies[q]);
	    }
	}
      // Next, do the same check for the products.
      for (int q = 0; q < (reactionDataVector[i]->pProductSpecies.size()); q++)
	{
	  bool ignoreSpecies = 0;
	  for (int e = 0; e < (reactionDataVector[i]->pSubstrateSpecies.size()); e++)
	    {
	      if (reactionDataVector[i]->pProductSpecies[q] == reactionDataVector[i]->pSubstrateSpecies[e])
		{
		  ignoreSpecies = 1;
		  break;
		}
	    }
	  if (!ignoreSpecies)
	    {
	      pSpeciesList.push_back (reactionDataVector[i]->pProductSpecies[q]);
	    }
	}
      // Time to build the reaction graph
      // First, include self once
      reactionDataVector[i]->pReactionGraph.push_back(reactionDataVector[i]); 
      //check for other reactions using the same species as substrates
      for (int t = 0; t < pSpeciesList.size(); t++)
	{
	  for (int p = 0; p < (reactionDataVector.size()); p++)
	    {
	      if (i != p) //ignore self
		{
		  for (int w = 0; w < (reactionDataVector[p]->
				       pSubstrateSpecies.size()); w++)
		    {
		      if (pSpeciesList[t] == 
			  reactionDataVector[p]->pSubstrateSpecies[w])
			{
			  exists = 0;
			  //check for previous listing of reaction
			  for (int o = 0; o < (reactionDataVector[i]->
					       pReactionGraph.size()); o ++)
			    if (reactionDataVector[i]->pReactionGraph[o] == 
				reactionDataVector[p])
			      {
				exists = 1;
				break;
			      }
			  if (!exists) //add if new reaction doesn't exist
			    {
			      reactionDataVector[i]->
				pReactionGraph.push_back(reactionDataVector[p]);
			      // once one reaction is added, no reason to check the rest
			      // of its substrates so, 
			      break;
			    }
			}
		    }
		}
	    }
	}
    }
  cout << "Finished!" << endl << endl;
}

void
reactionTypeA::dumpType (ostream& outFile, int index) const
{
  double h = 1;
  outFile << " = " << h << endl
	  << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}

void
reactionTypeB::dumpType (ostream& outFile, int index) const
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]);
  outFile << double (pSubstrateSpecies[0]->aAmountSpecies[index])  << "";
  outFile << " = " << h;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}


void
reactionTypeC::dumpType (ostream& outFile, int index) const
{
  double h1 = (pSubstrateSpecies[0]->aAmountSpecies[index]);
  double h2 = (pSubstrateSpecies[1]->aAmountSpecies[index]); 
  double h = h1*h2;
  
  outFile << h1 << "*"
	  << h2<< "";
  outFile << " = " << h;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
 
}


void
reactionTypeD::dumpType (ostream& outFile, int index) const
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[1]->aAmountSpecies[index]) - 1)/2;  
  outFile << double (pSubstrateSpecies[0]->aAmountSpecies[index])  << "*"
	  << double (pSubstrateSpecies[1]->aAmountSpecies[index]) << "-1)/2";
  outFile << " = " << h;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}

void
reactionTypeE::dumpType (ostream& outFile, int index) const
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    double (pSubstrateSpecies[1]->aAmountSpecies[index]) * 
    double (pSubstrateSpecies[2]->aAmountSpecies[index]);
  outFile << double (pSubstrateSpecies[0]->aAmountSpecies[index])  << "*"
	  << double (pSubstrateSpecies[1]->aAmountSpecies[index]) << "*"
	  << double (pSubstrateSpecies[2]->aAmountSpecies[index]) << "";
  outFile << " = " << h << endl;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}


void
reactionTypeFa::dumpType (ostream& outFile, int index) const
{ 
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[1]->aAmountSpecies[index]) - 1)/2 * 
    double (pSubstrateSpecies[2]->aAmountSpecies[index]);
  outFile << double (pSubstrateSpecies[0]->aAmountSpecies[index])  << "*"
	  << double (pSubstrateSpecies[1]->aAmountSpecies[index]) << "-1)/2*"
	  << double (pSubstrateSpecies[2]->aAmountSpecies[index]) << " ";
  outFile << " = " << h;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}

void
reactionTypeFb::dumpType (ostream& outFile, int index) const
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    double (pSubstrateSpecies[1]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[2]->aAmountSpecies[index]) - 1)/2;
  outFile << double (pSubstrateSpecies[0]->aAmountSpecies[index])  << "*"
	  << double (pSubstrateSpecies[1]->aAmountSpecies[index]) << "*"
	  << double (pSubstrateSpecies[2]->aAmountSpecies[index]) << "-1)/2";
  outFile << " = " << h;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}

void
reactionTypeG::dumpType (ostream& outFile, int index) const
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[1]->aAmountSpecies[index]) - 1) * 
    (double (pSubstrateSpecies[2]->aAmountSpecies[index]) - 2)/6;
  outFile << double (pSubstrateSpecies[0]->aAmountSpecies[index])  << "*"
	  << double (pSubstrateSpecies[1]->aAmountSpecies[index]) << "-1*"
	  << double (pSubstrateSpecies[2]->aAmountSpecies[index]) << "-2)/6";
  outFile << " = " << h;
  outFile << "*reaction rate: " << reactionRate << " = " << (h*reactionRate) << endl;
}

double
reactionTypeA::hcEqu(int index)
{
  double h = 1;
  return h*reactionRate;
}

double
reactionTypeB::hcEqu(int index)
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]);
          
       
  if (h<0)
    {
      cout << "h:  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp;
      cin >> tmp;
    }
  return h*reactionRate;
}

double
reactionTypeC::hcEqu(int index)
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * double (pSubstrateSpecies[1]->aAmountSpecies[index]); 
             
  if (h<0)
    {
      cout << "h:  " << h << endl;
      cout << "sub 1  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "sub 2  " << pSubstrateSpecies[1]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp;
      cin >> tmp;
    }
  return h*reactionRate;
}

double
reactionTypeD::hcEqu(int index)
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[1]->aAmountSpecies[index]) - 1)/2;  
      
  if (h<0)
    {
      cout << "h:  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[1]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp; 
      cin >> tmp;
    }
  return h*reactionRate;
}

double
reactionTypeE::hcEqu(int index)
{
 double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    double (pSubstrateSpecies[1]->aAmountSpecies[index]) * 
    double (pSubstrateSpecies[2]->aAmountSpecies[index]);
  if (h<0)
    {
      cout << "h:  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[1]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[2]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp;
      cin >> tmp;
    }
  return h*reactionRate;
}

double
reactionTypeFa::hcEqu(int index)
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[1]->aAmountSpecies[index]) - 1)/2 * 
    double (pSubstrateSpecies[2]->aAmountSpecies[index]);
  if (h<0)
    {
      cout << "h:  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[1]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[2]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp;
      cin >> tmp;
    }
  return h*reactionRate;
}

double
reactionTypeFb::hcEqu(int index)
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    double (pSubstrateSpecies[1]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[2]->aAmountSpecies[index]) - 1)/2;

  if (h<0)
    {
      cout << "h:  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[1]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[2]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp;
      cin >> tmp;
    }
return h*reactionRate;
}

double
reactionTypeG::hcEqu(int index)
{
  double h = double (pSubstrateSpecies[0]->aAmountSpecies[index]) * 
    (double (pSubstrateSpecies[1]->aAmountSpecies[index]) - 1) * 
    (double (pSubstrateSpecies[2]->aAmountSpecies[index]) - 2)/6;
  
  if (h<0)
    {
      cout << "h:  " << pSubstrateSpecies[0]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[1]->aAmountSpecies[index] << endl;
      cout << "h:  " << pSubstrateSpecies[2]->aAmountSpecies[index] << endl;
      cout << "Reaction rate:  " << reactionRate << endl;
      string tmp;
      cin >> tmp;
    }
  return h*reactionRate;
}

void 
reactionDataTable::initializeTimes()
{
  cout << "Initializing times . . .";
  for (int i = 0; i < (reactionDataVector.size()); i++)
    {
      for (int p = 0; p < replicate; p++)
	{
	  reactionDataVector[i]->times.push_back
	    (new timeStuff(reactionDataVector[i]->hcEqu(p)));
	}  
    }
  cout << "Finished!" << endl << endl;
}

void
timeStuff::calculateNewTime (double inHC, double totalTime)
{  
  oldhc = hc;
  hc = inHC;
  if (hc == 0)
    {
      timeInfinity = 1;
      time = FLT_MAX;
      
    }
  else 
    {
      timeInfinity = 0; 
      double rNum = 0;
      while ((rNum == 0) || (rNum == 1))
	rNum = 1-drand48();
      time = ((log (1/rNum))/hc)+totalTime; 
      //dumpTimeStuff(cout);
      
    }
  lastRxnNew = 1;
  
}

void
timeStuff::updateTime (double inHC, double totalTime)
{
  oldhc = hc;
  hc = inHC;
  if (hc == 0)
    {
      timeInfinity = 1;
      time = FLT_MAX;
    }
  else
    {
      timeInfinity = 0;
      time = ((oldhc/hc) * (time-totalTime))+totalTime;
      //dumpTimeStuff(cout);
      
    }
  lastRxnNew = 0;
  
}

void 
timeStuff::dumpTimeStuff(ostream& outFile)
{
  outFile << "Time: " << time;
  outFile << "\ttimeArrayPosition: " << timeArrayPosition;
  outFile << "\thc: " << hc;
  outFile << "\told hc: " << oldhc;
  outFile << "\ttimeInfinity: " << timeInfinity;
  outFile << "\tlast rxn new: " << lastRxnNew;
  outFile << endl;
  
}




