#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <cmath>
#include <iostream>

using namespace std;

#include "stoch.hh"

void 
species::setAmount(float newAmount, int arrayPosition)
    {
      aAmountSpecies[arrayPosition] = newAmount;
    }

speciesTable::~speciesTable(void)
{
  for(int i = 0; i < speciesVector.size(); i++)
    {
      delete speciesVector[i];
    }
}

species*
speciesTable::speciesLookup (const string& speciesName) const
{
  for (int i =0; i < speciesVector.size(); i++)
    {
      if (speciesName == speciesVector[i]->speciesName)
	return speciesVector[i];
    }
  cout << "Error!  Species " << speciesName << " not in species table!"  << endl;
  exit(0);
}

void 
speciesTable::addNewSpecies (species* pSpecies)
{
  speciesVector.push_back(pSpecies);
}

void 
speciesTable::readSpeciesFile (istream& speciesIn)
{
  cout << "Beginning to read species file . . ." << endl;
  cout << "\treplicates:  " << replicate << endl;
  while (!speciesIn.eof())
    { 
      //give name nothing so spaces before eol don't lead to data duplication
      int number = 0;
      string name;
      speciesIn >> name;
      
      //check to see if the species to be is going to be a switch dude.  If so, grab the 
      // name of the species that will be tracked and initialize the switch species.
      if (name == "switch")
	{
	  cout << "\tname:  " << name << endl;
	  name = "";
	  string comp1;
	  string comp2;
	  speciesIn >> comp1;
	  speciesIn >> comp2;
	  // set up the comparison for species A
	  species* tmp = speciesLookup(comp1);
	  tmp->comp = 1;
	  tmp->compSpecies = speciesLookup(comp2);
	  // set up the comparison for species B
	  tmp->compSpecies->comped = 1;
	  tmp->compSpecies->compedSpecies = tmp;
	  
	  
	  if (tmp->aAmountSpecies[0] > tmp->compSpecies->aAmountSpecies[0])
	    {
	      for (int i=0; i < replicate; i++)
		{
		  tmp->vGreater.push_back(1);
		}
	    }
	  else 
	    {
	      for (int i=0; i < replicate; i++)
  		{
		  tmp->vGreater.push_back(0);
		}
	    }
	  cout << "NAME: " << tmp->speciesName << endl;
	  cout << "comp: " << tmp->comp << endl;
	  cout << "comp name: " << tmp->compSpecies->speciesName << endl;
	}
      
		
	     
	  

      if (name.size() > 0)
  	{
	  speciesIn >> number;
	  addNewSpecies (new species(name, number, replicate));
	  cout << "\tname:  " << name << " amount:  " << number << endl;
	}
    }
  cout << "Finished!" << endl << endl;
}

void 
speciesTable::dumpSpeciesTable (ostream& outFile)
{
  for (int i=0; i < speciesVector.size(); i++)
    {
      outFile << "The species name is:  " << speciesVector[i]->speciesName 
	   <<"\tand for each replicate, the amount is:  ";
      for (int p = 0; p < replicate; p++)
	outFile << speciesVector[i]->aAmountSpecies[p] << " ";
      outFile << endl;
    }
  outFile << endl;
}

void
speciesTable::averator (ostream& outFile)
{
  for (int i=0; i < speciesVector.size(); i++)
    {
      if (speciesVector[i]->output)
	{
	  double average = 0;
	  for (int p = 0; p < replicate; p++)
	    {
	      average = average + speciesVector[i]->aAmountSpecies[p];
	    }
	  average = average/replicate;
	  double var = 0;
	  for (int p = 0; p < replicate; p++)
	    {
	      var = var + ((speciesVector[i]->aAmountSpecies[p] - average) * (speciesVector[i]->aAmountSpecies[p]-average));
	    }
	  var = var/(replicate-1);
	  double stdv = sqrt(var);
	  
	  outFile << (average-stdv) << "\t";
	  outFile << average << "\t";
	  outFile << (average+stdv) << "  \t";
	}
    }
}


