using namespace std;
#include <iostream>
class species
{
 public:
  std::string speciesName; 
  int replicate;
  bool output;
  double oldMean;
  double oldStndv;
  vector <float> aAmountSpecies;
  bool checkSpeciesAmount;
  int amountLimit;
  bool upperLimit;
  double lastTimeGtlt;
  double lastTimeEqui;
  
  
  
  // these two track spontaneous switching events.  The bool comp says whether or not
  // this species will be evaluated, the species* points to whom to compare against, 
  // and bool greater tracks if this species amount was greater than the one compared
  // to for the last comparison.

  bool comp;
  bool comped;
  
  species* compSpecies;
  species* compedSpecies;
  vector <bool> vGreater;
  
  bool lowLim;
  double timeLowLim;
  bool aveLim;
  double timeAveLim;
  bool highLim;
  double timeHighLim;
  
  species (const string& inName, float inAmount, int inReplicate)  :
    speciesName(inName),
    replicate(inReplicate),
    output(1),
    checkSpeciesAmount(0),
    lowLim(0),
    aveLim(0),
    highLim(0),
    comp (0),
    comped (0),
    lastTimeGtlt (0),
    lastTimeEqui (0)
    {
      for (int i=0; i < replicate; i++)
	aAmountSpecies.push_back (inAmount);
    }

  void setAmount (float newAmount, int arrayPosition);
};

class speciesTable
{
 public:
  int replicate;
  vector <species*> speciesVector;
  int steady;
  
  speciesTable (int inReplicate) :
    replicate(inReplicate),
    steady(0)
    {
    }
  ~speciesTable(void);

  void addNewSpecies (species* pSpecies);
  void readSpeciesFile (istream& speciesFileName);
  void dumpSpeciesTable (ostream& outFile);
  void averator (ostream& outFile);
  species* speciesLookup (const string& speciesName) const;
};

class timeStuff
{
 public:
  double time;
  int timeArrayPosition;
  double hc;
  double oldhc;
  bool lastRxnNew;
  
  
  bool timeInfinity;
  
  timeStuff (double inHC)
    {
      calculateNewTime (inHC,0);
    }
  
  void updateTime (double inHC, double totalTime);
  void calculateNewTime (double inHC, double totalTime); 
  void dumpTimeStuff (ostream& outFile);
};

class reactionData
{
 public:
  const string reactionName;
  vector <species*> pSubstrateSpecies;
  vector <species*> pProductSpecies;
  const float reactionRate;
  vector <reactionData*> pReactionGraph;
  vector <timeStuff*> times;
  
  reactionData (const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab);
  
  ~reactionData(void);
  
  virtual void dumpType(ostream& outFile, int index) const = 0;
  
  virtual double hcEqu(int index) = 0;
  
  void decreaseSubstrates (const int rep, const double totalTime, ostream& outFile);
  
  void increaseProducts (const int rep, const double totalTime, ostream& outFile);
};



class reactionTypeA : public reactionData
{
 public:
  reactionTypeA(const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type A." << endl;
    }

  void dumpType (ostream& outFile, int index) const;
 
  double hcEqu(int index);
};

class reactionTypeB : public reactionData
{
 public:
  reactionTypeB(const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)

    {
      cout << "\t" << inReactionName << ":  reaction type B." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};

class reactionTypeC : public reactionData
{
 public:
  reactionTypeC(const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type C." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};

class reactionTypeD : public reactionData
{
 public:
  reactionTypeD(const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type D." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};

class reactionTypeE : public reactionData
{
 public:
  reactionTypeE(const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type E." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};

class reactionTypeFa : public reactionData
{
 public:
  reactionTypeFa(const string inReactionName, vector <string> inSubstrates, 
		 vector <string> inProducts, 
		 const float inReactionRate, 
		 const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type Fa." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};   

class reactionTypeFb : public reactionData
{
 public:
  reactionTypeFb(const string inReactionName, vector <string> inSubstrates, 
		 vector <string> inProducts, 
		 const float inReactionRate, 
		 const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type Fb." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};   

class reactionTypeG : public reactionData
{
 public:
  reactionTypeG(const string inReactionName, vector <string> inSubstrates, 
		vector <string> inProducts, 
		const float inReactionRate, 
		const speciesTable& specTab) : 
    reactionData (inReactionName, inSubstrates, 
		  inProducts, 
		  inReactionRate, 
		  specTab)
    {
      cout << "\t" << inReactionName << ":  reaction type G." << endl;
    }
  void dumpType (ostream& outFile, int index) const;

  double hcEqu(int index);
};   

class reactionDataTable 
{
 
  
 public:

  int replicate;
  reactionDataTable (int inReplicate) :
    replicate(inReplicate)
    {}
  
  vector <reactionData*> reactionDataVector;

  ~reactionDataTable(void);

  void addNewReaction (reactionData* pReactionData);

  void readReactionDataFile (istream& rxnIn, const speciesTable& rSpecTab);

  void dumpReactionDataTable (ostream& outFile);

  int getType (vector <string> subs);

  void makeReactionGraph ();
  
  void initializeTimes ();

 
};

class prioriQ
{
 public:

  reactionData** timeArray;
  int sizeTimeArray;
  double totalTime;
   
  prioriQ (reactionDataTable& rReactionDataTable, int index);
  void dumpPrioriQ (int index,ostream& outFile);
  void dumpPrioriQ0 (int index, ostream& outFIle);
  
  void switchTimes (const int p, const int q, int index);
  int findParent (int node);
  void bubbleUp (int node, int index);
  int childCheck (int node);
  void bubbleDown (int node, int index);
  void bubble (int node, int index);
  void executeReaction (int node, int index, ostream& outFile);
  void updatePQ (double time, int index);
  

};

class prioriQtable
{
 public:
  vector <prioriQ*> pqVector;
  int replicate;
  
  prioriQtable (int inReplicate, reactionDataTable& rReactionDataTable);
  ~prioriQtable (void);
  void dumpPrioriQtable (ostream& outFile);
  
};
  









